/********
Animator Class
*********/
#include "animator.h"
#include <iostream>
#include <queue>

using namespace std;

Animator::Animator() {
	currFrame = NULL;
	length = 0;
}

Animator::~Animator() {
	while( !textures.empty() ) {
		pTexture* ptr = textures.back();
		textures.pop_back();
		delete ptr;
	}
	while( !materials.empty() ) {
		pMaterial* ptr = materials.back();
		materials.pop_back();
		delete ptr;
	}
	while( !models.empty() ) {
		pModel* ptr = models.back();
		models.pop_back();
		delete ptr;
	}
	while( !frames.empty() ) {
		Frame* ptr = frames.back();
		frames.pop_back();
		delete ptr;
	}
	while( !psystem.empty() ) {
		ParticleSystem* ptr = psystem.back();
		psystem.pop_back();
		delete ptr;
	}
}

// Parse Animation portion of XML file
bool Animator::loadScene( TiXmlDocument* doc ) {
	TiXmlElement* element = NULL;
	if( ! doc->LoadFile() ) {
		return false;
	}
	element = doc->FirstChildElement( "texture" ); 
	while( element != NULL ){
		const char * tname = element->Attribute("name");
		TiXmlElement* fileElem = element->FirstChildElement( "file" );
		TiXmlText* fileText = fileElem->FirstChild()->ToText();
		pTexture *tex = new pTexture( tname, fileText->Value() );
		addTexture( tex );
		element = element->NextSiblingElement("texture");
	}
	element = doc->FirstChildElement( "material" ); 

	while( element != NULL ){
		const char * mname = element->Attribute("name");
		pMaterial *mat = new pMaterial( mname );
		TiXmlElement* diffElem = element->FirstChildElement( "diffuse" );
		if( diffElem ) {
			rgb color;
			int r, g, b;
			TiXmlText* colorText = diffElem->FirstChild()->ToText();
			sscanf( colorText->Value(), "%d,%d,%d", &r, &g, &b);
			color.r = (Uint8)r;
			color.g = (Uint8)g;
			color.b = (Uint8)b;
			mat->setDiffuse( color );
		}
		TiXmlElement* trefElem = element->FirstChildElement( "tref" );
		while( trefElem != NULL ) {
			int num;
			trefElem->Attribute("num", &num);
			TiXmlText* trefText = trefElem->FirstChild()->ToText();
			pTexture *tex = getTexture( trefText->Value() );
			if( tex != NULL ) mat->setTexture( num, tex );
			trefElem = trefElem->NextSiblingElement("tref");
		}
		addMaterial( mat );
		element = element->NextSiblingElement("material");
	}
	element = doc->FirstChildElement( "object" ); 
	while( element != NULL ){
		bool smooth = false;;
		const char *oname = element->Attribute("name");
		pModel *obj = new pModel( oname );
		TiXmlElement* fileElem = element->FirstChildElement( "file" );
		TiXmlText* fileText = fileElem->FirstChild()->ToText();
		TiXmlElement* matElem = element->FirstChildElement( "mref" );
		if( matElem ) {
			TiXmlText* matText = matElem->FirstChild()->ToText();
			pMaterial *mat = getMaterial( matText->Value() );
			obj->setMaterial( mat );
		}
		TiXmlElement* polyElem = element->FirstChildElement( "poly" );
		if( polyElem ) {
			TiXmlText* polyText = polyElem->FirstChild()->ToText();
			if( strcmp( "smooth", polyText->Value() ) == 0 ) 
				smooth = true;
		}
		TiXmlElement* tmapElem = element->FirstChildElement( "tmap" );
		if( tmapElem ) {
			const char* axis = tmapElem->Attribute("axis");
			if( axis != NULL ) {
				int a;
				if( axis[0] == '-' ) {
					if( axis[1] == 'x' ) a = 1;
					if( axis[1] == 'z' ) a = 3;
					if( axis[1] == 'y' ) a = 5;
				} else {
					if( axis[0] == 'x' ) a = 0;
					if( axis[0] == 'z' ) a = 2;
					if( axis[0] == 'y' ) a = 4;
				}
				obj->setAxis( a );
			}
			Point3D pos;
			TiXmlText* tmapText = tmapElem->FirstChild()->ToText();
			sscanf( tmapText->Value(), "%le,%le,%le", &pos[0], &pos[1], &pos[2]);
			obj->setTMap( pos );
		}
		if( obj->loadOBJ( fileText->Value() ) )
			obj->genDrawList( smooth );
		addModel( obj );
		element = element->NextSiblingElement("object");
	}
	loadFrames( doc->FirstChildElement( "keyframes" ) ); 
	
	return false;
}

// Parse Keyframe section of XML file
void Animator::loadFrames( TiXmlElement * element ) {
	if( element == NULL ) return;
	TiXmlElement *frameElem;
	frameElem = element->FirstChildElement( "frame" );
	while( frameElem != NULL ) {
		int time;
		frameElem->Attribute( "time", &time );
		Frame * frame = new Frame( time );
		TiXmlElement* orefElem = frameElem->FirstChildElement( "oref" );
		while( orefElem != NULL ) {
			const char* oname = orefElem->Attribute( "name" );
			pModel* obj = getModel( oname );
			if( obj != NULL ) {
				ModelFrame *mf = new ModelFrame( obj );
				TiXmlElement* posElem = orefElem->FirstChildElement( "pos" );
				if( posElem ) {
					Point3D pos;
					TiXmlText* posText = posElem->FirstChild()->ToText();
					sscanf( posText->Value(), "%le,%le,%le", &pos[0], &pos[1], &pos[2]);
					mf->setPos( pos );
				}
				TiXmlElement* rotElem = orefElem->FirstChildElement( "rot" );
				while( rotElem != NULL ) {
					const char* axis = rotElem->Attribute("axis");
                    TiXmlText* rotText = rotElem->FirstChild()->ToText();
					double angle;
					sscanf( rotText->Value(), "%le", &angle );
					if( axis[0] == 'x' ) mf->setRX( angle );
					else if( axis[0] == 'y' ) mf->setRY( angle );
					else if( axis[0] == 'z' ) mf->setRZ( angle );
					rotElem = rotElem->NextSiblingElement("rot");
				}
				frame->addModelFrame( mf );
			}
			orefElem = orefElem->NextSiblingElement( "oref" );
		}
		addFrame( frame );
		frameElem = frameElem->NextSiblingElement( "frame" );
	}
	TiXmlElement* particleElem = element->FirstChildElement( "particle" );
	while( particleElem != NULL ) {
		int time;
		Point3D pos;
		Vector3D vec;
		rgb color = RGB_0;
		double accel = 0;
		particleElem->Attribute( "time", &time );
		ParticleSystem * ps = new ParticleSystem( time );
		TiXmlElement* posElem = particleElem->FirstChildElement( "pos" );
		if( posElem ) {
			TiXmlText* posText = posElem->FirstChild()->ToText();
			sscanf( posText->Value(), "%le,%le,%le", &pos[0], &pos[1], &pos[2]);
		}
		TiXmlElement* vecElem = particleElem->FirstChildElement( "velocity" );
		if( vecElem ) {
			TiXmlText* vecText = vecElem->FirstChild()->ToText();
			sscanf( vecText->Value(), "%le,%le,%le", &vec[0], &vec[1], &vec[2]);
		}
		TiXmlElement* colElem = particleElem->FirstChildElement( "color" );
		if( colElem ) {
			int r, g, b;
			TiXmlText* colorText = colElem->FirstChild()->ToText();
			sscanf( colorText->Value(), "%d,%d,%d", &r, &g, &b);
			color.r = (Uint8)r;
			color.g = (Uint8)g;
			color.b = (Uint8)b;
		}
		TiXmlElement* accelElem = particleElem->FirstChildElement( "accel" );
		if( accelElem ) {
			TiXmlText* accelText = accelElem->FirstChild()->ToText();
			sscanf( accelText->Value(), "%le", &accel);
		}
		ps->initialize( pos, vec, accel, color );
		addParticle( ps );
		particleElem = particleElem->NextSiblingElement( "particle" );
	}

}

// dummy load Scene for testing purposes
void Animator::loadScene( ){
	rgb diffuse = { 0, 128, 255 };
	pMaterial * mat = new pMaterial( "red", diffuse );
	materials.push_back( mat );
	pModel * mod = new pModel( "cow" );
	mod->loadOBJ( "cow.obj" );
	mod->setMaterial( mat );
	mod->genDrawList( true );
	models.push_back( mod );
}

void Animator::render( int t, bool rt ) {
	if( !frames.empty() ) {
		if( currFrame == NULL || t < currFrame->getTime() ) {
			currFrame = NULL;
			for( unsigned int i = 1; i < frames.size(); i++ ) {
				if( frames[i]->getTime() > t ) {
					currFrame = frames[i-1];
					break;
				} 
			}
			if( currFrame == NULL ) currFrame = frames.back();
		} else {
			Frame *nextFrame = currFrame->getNextFrame();
			while( nextFrame != NULL ) {
				if( nextFrame->getTime() > t ) break;
				currFrame = nextFrame;
				nextFrame = currFrame->getNextFrame();
			}
		}
		if( !rt ) {
			glPushAttrib( GL_ALL_ATTRIB_BITS );
			glDisable(GL_TEXTURE_2D);
		}
		currFrame->render( t, rt );
		if( !rt ) {
			glPopAttrib();
		}
	}
	if( !psystem.empty() ) {
		 prq drawp;

		glPushAttrib( GL_ALL_ATTRIB_BITS );
		glEnable( GL_BLEND );

		for( unsigned int i = 0; i < psystem.size(); i++ ) {
			if( psystem[i]->getStart() > t ) continue;
			if( psystem[i]->getEnd() >= t ) {
				psystem[i]->render( t, drawp );
			}
		}
		while( !drawp.empty() ) {
			PRef pr = drawp.top();
			drawp.pop();
			pr.p->render( t );
		}

		glFlush();
		glPopAttrib();
	}
}

void Animator::addTexture( pTexture * tex ) {
	textures.push_back( tex );
}

pTexture *Animator::getTexture( const char * name ) {
	for( unsigned int i = 0; i < textures.size(); i++ ) {
		if( textures[i]->getname().compare( name ) == 0 ) {
			return textures[i];
		}
	}
	return NULL;
}

pMaterial *Animator::getMaterial( const char * name ) {
	for( unsigned int i = 0; i < materials.size(); i++ ) {
		if( materials[i]->getname().compare( name ) == 0 ) {
			return materials[i];
		}
	}
	return NULL;
}

pModel *Animator::getModel( const char * name ) {
	for( unsigned int i = 0; i < models.size(); i++ ) {
		if( models[i]->getname().compare( name ) == 0 ) {
			return models[i];
		}
	}
	return NULL;
}

void Animator::addMaterial( pMaterial * mat ){
	materials.push_back( mat );
}

void Animator::addModel( pModel * obj ){
	models.push_back( obj );
}

void Animator::addFrame( Frame * frame ){
	if( frames.empty() ) {
		frames.push_back( frame );
	} else if( frames.back()->getTime() < frame->getTime() ) {
		frames.back()->setNextFrame( frame );
		frames.push_back( frame );
	}
	if( frame->getTime() > length ) {
		length = frame->getTime();
	}
}

void Animator::addParticle( ParticleSystem * ps ){
	if( psystem.empty() ) {
		psystem.push_back( ps );
	} else if( psystem.back()->getStart() <= ps->getStart() ) {
		psystem.push_back( ps );
	}
	if( ps->getEnd() > length ) {
		length = ps->getEnd();
	}
}
